/*
 * Realtek RT8181 gpio driver, quick hack
 *
 * Copyright (C) 2004 Jeroen Domburg (Sprite_tm)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 */

/*
Uses /dev/gpio, character device, major 10, minor 123

Read: always read 12 bytes. Seek is ignored.
0: word: GPIO data pins
4: word: GPIO data direction
8: word: GPIO last written

Write: 
Seek to the appropriate position.
Only write one word. Word 0 and 4 are writable. Anything else will give an 
error.

*/

#include <linux/config.h>
#include <linux/version.h>
#include <linux/fs.h>
#include <linux/string.h>
#include <linux/module.h>
#include <linux/miscdevice.h>
#include <linux/init.h>
#include <linux/miscdevice.h>
#include <asm/uaccess.h>
#include <asm/rtl8181.h>

#define rt8181_gpio_MINOR 123


//Because the data register doesn't seem to be read-modify-write-compatible,
//we'll keep track of the last value written ourselves.
long rtl8181_gpio_sent;

//static int rt8181_gpio_open(struct inode *inode, struct file *file);
static int rt8181_gpio_release(struct inode *inode, struct file *file);
static ssize_t rt8181_gpio_read(struct file *filp, char *buffer, size_t length, loff_t *offset);
static ssize_t rt8181_gpio_write(struct file *filp, const char *buffer, size_t length, loff_t *offset);

static struct file_operations rt8181_gpio_fops = {
	owner:		THIS_MODULE,
	read:		rt8181_gpio_read,
	write:		rt8181_gpio_write,
//	open:		rt8181_gpio_open,
	release:	rt8181_gpio_release
};

static struct miscdevice rt8181_gpio_miscdev = {
	//register minor
	rt8181_gpio_MINOR,
	"rt8181_gpio",
	&rt8181_gpio_fops
};


static ssize_t rt8181_gpio_read(struct file *filp, char *buffer, size_t length, loff_t *offset) {
    unsigned long buff[3];
    char *tmp=buffer;
    int i;

    //Fill buffer with appropriate values
    buff[0]=rtl_inl(0x44);
    buff[1]=rtl_inl(0x40);
    buff[2]=rtl8181_gpio_sent;

//    printk("gpio.o: Read: %lx %lx %lx\n",buff[0],buff[1],buff[2]);

    //and copy to userspace    
    for (i = 0; length-- > 0 && i < 4; ++i) {
        put_user((long)buff[i], (long *) tmp); 
	tmp+=4;
    }

    return (tmp-buffer);
}

static ssize_t rt8181_gpio_write(struct file *filp, const char *buffer, size_t length, loff_t *offset){
    unsigned long buff;
    if (length!=4) return -EFAULT; //Only longs are supported.
    copy_from_user(&buff,buffer,4);
    if (*offset==0) { //Set output
//	printk("gpio.o: Set output to %lx\n",buff);
	rtl_outl(0x44,buff);
	rtl8181_gpio_sent=buff;
	return 4;
    }
    if (*offset==4) { //Set data direction
//	printk("gpio.o: Set dir to %lx\n",buff);
	rtl_outl(0x40,buff);
	return 4;
    }
    return -EFAULT;
}

static int
rt8181_gpio_release(struct inode *inode, struct file *file)
{
	return 0;
}


static __init int rt8181_gpio_init(void)
{
	misc_register(&rt8181_gpio_miscdev);
	printk("rtl8181 gpio module loaded.\n");
	return 0;
}

static __exit void rt8181_gpio_exit(void)
{
	misc_deregister(&rt8181_gpio_miscdev);
	printk("rtl8181 gpio module unloaded.\n");
}

module_init(rt8181_gpio_init);
module_exit(rt8181_gpio_exit);


MODULE_AUTHOR ("Jeroen Domburg");
MODULE_LICENSE("GPL");
